/* -*- mode: c; c-file-style: "linux"; -*- */
/*
Copyright (C) 2003 Robert Ham <rah@bash.sh>
Copyright (C) 2001 Paul Davis

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Id: net_driver.c,v 1.10 2005/12/11 17:16:18 torbenh Exp $
*/


#include <math.h>
#include <stdio.h>
#include <memory.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/poll.h>

#include <jack/types.h>
//#include <jack/internal.h>
#include <jack/engine.h>
#include <sysdeps/time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <endian.h>

#include "net_driver.h"

#undef DEBUG_WAKEUP

static int
net_driver_sync_cb( jack_transport_state_t state, jack_position_t *pos, net_driver_t *driver ) {
    return driver->sync_state;
}


static jack_nframes_t 
net_driver_wait (net_driver_t *driver, int extra_fd, int *status,
	float *delayed_usecs)
{
    // ok... we wait for a packet on the socket
    // TODO: 
    // we should be able to run freely if the sync source is not transmitting
    // especially we should be able to compensate for packet loss somehow.
    // but lets try this out first.

    fd_set read_fds;
    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 100000;
    int res = 0;
    while ( res == 0)
    {
	FD_ZERO(&read_fds);
	FD_SET(driver->sockfd, &read_fds);
	res = select(FD_SETSIZE, &read_fds, NULL,NULL,&tv); // wait here until there is a packet to get
    }

    //driver->srcaddress_valid = 0;

    driver->last_wait_ust = jack_get_microseconds ();
    driver->engine->transport_cycle_start (driver->engine,
	    driver->last_wait_ust);

    /* this driver doesn't work so well if we report a delay */
    *delayed_usecs = 0;		/* lie about it */
    *status = 0;
    return driver->period_size;
}

static inline int
net_driver_run_cycle (net_driver_t *driver)
{
	//puts ("net_driver_run_cycle");
	//printf("c");
	jack_engine_t *engine = driver->engine;
	int wait_status;
	float delayed_usecs;

	jack_nframes_t nframes = net_driver_wait (driver, -1, &wait_status,
						&delayed_usecs);
	if (nframes == 0) {
		/* we detected an xrun and restarted: notify
		* clients about the delay. */
		engine->delay (engine, delayed_usecs );
		return 0;
	} 

	if (wait_status == 0)
		return engine->run_cycle (engine, nframes, delayed_usecs);

	if (wait_status < 0)
		return -1;
	else
		return 0;
}

static int
net_driver_null_cycle (net_driver_t* driver, jack_nframes_t nframes)
{
    // XXX: ok... this is sort of what needs te be done...
    // send and receive null packets...
    // grr... need to extract the packet functions....
    //
    // FIXME: only recives a packet. but thats enough for the fix...
    //
    // i saw a higher latency after several minutes of running. i dont see how this could be related
    // to not sending out a packet... but you never know...

	socklen_t address_size = sizeof(struct sockaddr_in);
	int bufsize =  sizeof(jack_default_audio_sample_t) * driver->capture_channels * nframes + sizeof(jacknet_packet_header);
	unsigned int *packet_buf;

	packet_buf = alloca( bufsize );

	// XXX: fill out packet header
	//jacknet_packet_header *pkthdr = packet_buf;
	
	if( !driver->srcaddress_valid ) {
		puts("!driver->srcaddress_valid");
		return 0;	
	}
	recvfrom( driver->sockfd, packet_buf, bufsize, 
		MSG_WAITALL, (struct sockaddr*) &driver->syncsource_address, &address_size );
	return 0;
}

static int
net_driver_bufsize (net_driver_t* driver, jack_nframes_t nframes)
{
	if (nframes < 64)
		return EINVAL;

	driver->period_size = nframes;  
	driver->period_usecs = 
		(jack_time_t) floor ((((float) nframes) / driver->sample_rate)
				* 1000000.0f);

	/* tell the engine to change its buffer size */
	driver->engine->set_buffer_size (driver->engine, nframes);

	return 0;
}

static int
net_driver_read (net_driver_t* driver, jack_nframes_t nframes)
{
	jack_default_audio_sample_t* buf;
	jack_port_t *port;
	socklen_t address_size = sizeof(struct sockaddr_in);
	jack_position_t local_trans_pos;
	jack_transport_state_t local_trans_state;
	int bufsize =  sizeof(jack_default_audio_sample_t) * driver->capture_channels * nframes + sizeof(jacknet_packet_header);
	unsigned int *packet_buf, *packet_bufX;

	packet_buf = alloca( bufsize );

	jacknet_packet_header *pkthdr = (jacknet_packet_header *)packet_buf;

	
	if( !driver->srcaddress_valid ) {
		puts("!driver->srcaddress_valid");
		return 0;	
	}
	int len = recvfrom( driver->sockfd, packet_buf, bufsize, 
		MSG_WAITALL, (struct sockaddr*) &driver->syncsource_address, &address_size );

//	while( (recvfrom( driver->sockfd, packet_buf, bufsize, 
//		MSG_WAITALL|MSG_DONTWAIT, (struct sockaddr*) &driver->syncsource_address, &address_size )) > 0 ) ;

	if (len < bufsize) {
	    printf( "buf_too_small len=%d, bufsize=%d\n", len, bufsize );
		//puts("too small buffer");
		return 0;
	}
	packet_bufX = packet_buf + sizeof(jacknet_packet_header) / sizeof(jack_default_audio_sample_t);

	
	// read local transport info....
	local_trans_state = jack_transport_query( driver->client, &local_trans_pos );

	// compare to the data we get from the master machine...
	// and doTheRightThing(tm)
	
	    
	// Now check if we have to start or stop local transport to sync to remote...
	switch( pkthdr->transport_state ) {
	    case JackTransportStarting:
		// the master transport is starting... so we set our reply to the sync_callback;
		driver->sync_state = FALSE;
		break;
	    case JackTransportStopped:
		driver->sync_state = TRUE;
		if( local_trans_pos.frame != (pkthdr->jack_pos.frame) ) {
		    jack_transport_locate( driver->client, (pkthdr->jack_pos.frame ) );
		    printf( "locate to %d\n", pkthdr->jack_pos.frame + nframes);
		}
		if( local_trans_state != JackTransportStopped )
		    jack_transport_stop( driver->client );
		break;
	    case JackTransportRolling:
		driver->sync_state = TRUE;
//		if( local_trans_pos.frame != (pkthdr->jack_pos.frame + 2*nframes ) ) {
//		    jack_transport_locate( driver->client, (pkthdr->jack_pos.frame + nframes ) );
//		    printf( "locate to %d\n", pkthdr->jack_pos.frame + nframes);
//		}
		if( local_trans_state != JackTransportRolling )
		    jack_transport_start( driver->client );
		break;

	    case JackTransportLooping:
		break;
	}
	

	    
	channel_t chn = 0;
	JSList *node = driver->capture_ports;
	while ( node != NULL)
	{
		port = (jack_port_t *) node->data;
		buf = jack_port_get_buffer (port, nframes);
	
		int i;
		unsigned int val;
		for (i = 0; i < nframes; i++)
		{
			unsigned int *ptr = (unsigned int *)&packet_bufX[i];
			val = ntohl( *ptr);
			buf[i] = *(float *)&val;
		}
		packet_bufX = (packet_bufX + nframes);
		node = jack_slist_next (node);
		chn++;
	}
	return 0;
}

static int
net_driver_write (net_driver_t* driver, jack_nframes_t nframes)
{
    static int framecnt = 0;

    jack_default_audio_sample_t *buf;
    jack_port_t *port;
    unsigned int *packet_buf, *packet_bufX;

    int packet_size = sizeof(jack_default_audio_sample_t) * driver->playback_channels * nframes + sizeof(jacknet_packet_header);

    packet_buf = alloca( packet_size );

    jacknet_packet_header *pkthdr = (jacknet_packet_header *)packet_buf;

    // offset packet_bufX by the packetheader.
    packet_bufX = packet_buf + sizeof(jacknet_packet_header) / sizeof(jack_default_audio_sample_t);

    pkthdr->sync_state = (driver->engine->control->sync_remain <= 1);
    pkthdr->frame_cnt = (framecnt++);

    channel_t chn = 0;
    JSList *node = driver->playback_ports;
    while ( node != NULL)
    {
	port = (jack_port_t *) node->data;
	buf = jack_port_get_buffer (port, nframes);

	int i;
	unsigned int val;
	for (i=0; i < nframes; i++) {
	    unsigned int *ptr = (unsigned int *)&buf[i];
	    val = htonl( *ptr);
	    packet_bufX[i]=val;
	}

	//		unsigned int *ptr = (unsigned int *)&buf[12];
	//		printf("sent=%X, val=%d\n", packet_bufX[12], val);
	//		//printf("packet_bufX[1]=%f, buf[3]=%f \n",packet_bufX[1], buf[3]);

	packet_bufX = (packet_bufX + nframes);
	node = jack_slist_next (node);
	chn++;
    }

    if( driver->srcaddress_valid )
	sendto( driver->outsockfd, packet_buf, packet_size, 
		0, (struct sockaddr*)&driver->syncsource_address, sizeof(struct sockaddr_in) ); 

    return 0;
}


static int
net_driver_attach (net_driver_t *driver)
{
puts ("net_driver_attach");
	jack_port_t * port;
	char buf[32];
	unsigned int chn;
	int port_flags;

	driver->engine->set_buffer_size (driver->engine, driver->period_size);
	driver->engine->set_sample_rate (driver->engine, driver->sample_rate);

	jack_set_sync_callback( driver->client, (JackSyncCallback) net_driver_sync_cb, driver );

	port_flags = JackPortIsOutput|JackPortIsPhysical|JackPortIsTerminal;

//#if 0
	for (chn = 0; chn < driver->capture_channels; chn++)
	{
		snprintf (buf, sizeof(buf) - 1, "capture_%u", chn+1);

		port = jack_port_register (driver->client, buf,
					JACK_DEFAULT_AUDIO_TYPE,
					port_flags, 0);
		if (!port)
		{
			jack_error ("DUMMY: cannot register port for %s", buf);
			break;
		}

		driver->capture_ports =
			jack_slist_append (driver->capture_ports, port);
	}
//#endif	
	port_flags = JackPortIsInput|JackPortIsPhysical|JackPortIsTerminal;

	for (chn = 0; chn < driver->playback_channels; chn++)
	{
		snprintf (buf, sizeof(buf) - 1, "playback_%u", chn+1);

		port = jack_port_register (driver->client, buf,
					JACK_DEFAULT_AUDIO_TYPE,
					port_flags, 0);

		if (!port)
		{
			jack_error ("DUMMY: cannot register port for %s", buf);
			break;
		}

		driver->playback_ports =
			jack_slist_append (driver->playback_ports, port);
	}

	jack_activate (driver->client);

	return 0;
}

static int
net_driver_detach (net_driver_t *driver)
{
	JSList * node;

	if (driver->engine == 0)
		return 0;
//#if 0
	for (node = driver->capture_ports; node; node = jack_slist_next (node))
		jack_port_unregister (driver->client,
				((jack_port_t *) node->data));

	jack_slist_free (driver->capture_ports);
	driver->capture_ports = NULL;
//#endif
		
	for (node = driver->playback_ports; node; node = jack_slist_next (node))
		jack_port_unregister (driver->client,
				((jack_port_t *) node->data));

	jack_slist_free (driver->playback_ports);
	driver->playback_ports = NULL;

	return 0;
}


static void
net_driver_delete (net_driver_t *driver)
{
	jack_driver_nt_finish ((jack_driver_nt_t *) driver);
	free (driver);
}

static jack_driver_t *
net_driver_new (jack_client_t * client,
		char *name,
		unsigned int capture_ports,
		unsigned int playback_ports,
		jack_nframes_t sample_rate,
		jack_nframes_t period_size,
		unsigned int listen_port)
{
	net_driver_t * driver;
	struct sockaddr_in address;

	printf ("creating net driver ... %s|%" PRIu32 "|%" PRIu32
		"|%u|%u|%u\n", name, sample_rate, period_size, listen_port,
		capture_ports, playback_ports);

driver = (net_driver_t *) calloc (1, sizeof (net_driver_t));

	jack_driver_nt_init ((jack_driver_nt_t *) driver);

	driver->write         = (JackDriverWriteFunction)      net_driver_write;
	driver->read          = (JackDriverReadFunction)       net_driver_read;
	driver->null_cycle    = (JackDriverNullCycleFunction)  net_driver_null_cycle;
	driver->nt_attach     = (JackDriverNTAttachFunction)   net_driver_attach;
	driver->nt_detach     = (JackDriverNTDetachFunction)   net_driver_detach;
	driver->nt_bufsize    = (JackDriverNTBufSizeFunction)  net_driver_bufsize;
	driver->nt_run_cycle  = (JackDriverNTRunCycleFunction) net_driver_run_cycle;

	driver->period_usecs =
		(jack_time_t) floor ((((float) period_size) / sample_rate)
				* 1000000.0f);
	driver->sample_rate = sample_rate;
	driver->period_size = period_size;
	driver->listen_port   = listen_port;
	driver->last_wait_ust = 0;

	driver->capture_channels  = capture_ports;
	driver->capture_ports     = NULL;
	driver->playback_channels = playback_ports;
	driver->playback_ports    = NULL;

	driver->client = client;
	driver->engine = NULL;

	driver->sockfd = socket( PF_INET, SOCK_DGRAM, 0 );
	if( driver->sockfd == -1 ) {
	printf( "socket error\n" );
	return NULL;
	}
	address.sin_family = AF_INET;
	address.sin_port = htons( driver->listen_port );
	address.sin_addr.s_addr = htonl( INADDR_ANY );
	if( bind( driver->sockfd, (struct sockaddr *) &address, sizeof(address) ) < 0 ) {
	printf( "bind error\n" );
	return NULL;
	}

	driver->outsockfd = socket( PF_INET, SOCK_DGRAM, 0 );
	if( driver->sockfd == -1 ) {
	printf( "socket error\n" );
	return NULL;
	}
	driver->srcaddress_valid = 0;
	
	char msg;
	socklen_t address_size = sizeof(struct sockaddr_in);
	recvfrom( driver->sockfd, &msg, sizeof(char), 0, (struct sockaddr*)&driver->syncsource_address, &address_size );
	driver->srcaddress_valid = 1;

	return (jack_driver_t *) driver;
}


/* DRIVER "PLUGIN" INTERFACE */

jack_driver_desc_t *
driver_get_descriptor ()
{
	jack_driver_desc_t * desc;
	jack_driver_param_desc_t * params;
	unsigned int i;

	desc = calloc (1, sizeof (jack_driver_desc_t));
	strcpy (desc->name, "net");
	desc->nparams = 5;

	params = calloc (desc->nparams, sizeof (jack_driver_param_desc_t));

	i = 0;
	strcpy (params[i].name, "capture");
	params[i].character  = 'C';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 2U;
	strcpy (params[i].short_desc, "Number of capture ports");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "playback");
	params[i].character  = 'P';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 2U;
	strcpy (params[i].short_desc, "Number of playback ports");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "rate");
	params[i].character  = 'r';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 48000U;
	strcpy (params[i].short_desc, "Sample rate");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "period");
	params[i].character  = 'p';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 1024U;
	strcpy (params[i].short_desc, "Frames per period");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "listenport");
	params[i].character  = 'l';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 3000U;
	strcpy (params[i].short_desc,
		"The Port we are listening on for sync packets");
	strcpy (params[i].long_desc, params[i].short_desc);

	desc->params = params;

	return desc;
}

const char driver_client_name[] = "net_pcm";

jack_driver_t *
driver_initialize (jack_client_t *client, const JSList * params)
{
	jack_nframes_t sample_rate = 48000;
	jack_nframes_t period_size = 1024;
	unsigned int capture_ports = 2;
	unsigned int playback_ports = 2;
	unsigned int listen_port = 3000;
	const JSList * node;
	const jack_driver_param_t * param;

	for (node = params; node; node = jack_slist_next (node)) {
		param = (const jack_driver_param_t *) node->data;

		switch (param->character) {

		case 'C':
		capture_ports = param->value.ui;
		break;

		case 'P':
		playback_ports = param->value.ui;
		break;

		case 'r':
		sample_rate = param->value.ui;
		break;

		case 'p':
		period_size = param->value.ui;
		break;

		case 'l':
		listen_port = param->value.ui;
		break;
				
		}
	}


	return net_driver_new (client, "net_pcm", capture_ports,
				playback_ports, sample_rate, period_size,
				listen_port);
}

void
driver_finish (jack_driver_t *driver)
{
	net_driver_delete ((net_driver_t *) driver);
}
