/** @file simple_client.c
 *
 * @brief This simple client demonstrates the basic features of JACK
 * as they would be used by many applications.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <netinet/in.h>
#include <netdb.h>

#include <alloca.h>

#include <jack/jack.h>
#include <endian.h>

#include "net_driver.h"

//jack_port_t *input_port1, *input_port2;
//jack_port_t *output_port1, *output_port2;
JSList	   *capture_ports = NULL;
int capture_channels = 2;
JSList	   *playback_ports = NULL;
int playback_channels = 2;

jack_client_t *client;

int outsockfd;
int insockfd;
struct sockaddr destaddr;
struct sockaddr bindaddr;


int recv_channels;
int recv_smaple_format;
int sync_state;


/*
 * This Function allocates all the I/O Ports which are added the lists.
 * XXX: jack-midi is underway... so dont forget it.
 */

void alloc_ports( int n_capture, int n_playback ) {

    int port_flags = JackPortIsOutput;
    int chn;
    jack_port_t *port;
    char buf[32];

    capture_ports = NULL;
    for (chn = 0; chn < n_capture; chn++)
    {
	snprintf (buf, sizeof(buf) - 1, "capture_%u", chn+1);

	port = jack_port_register (client, buf,
		JACK_DEFAULT_AUDIO_TYPE,
		port_flags, 0);

	if (!port)
	{
	    printf( "jacknet_client: cannot register port for %s", buf);
	    break;
	}

	capture_ports = jack_slist_append (capture_ports, port);
    }

    port_flags = JackPortIsInput;

    playback_ports = NULL;
    for (chn = 0; chn < n_playback; chn++)
    {
	snprintf (buf, sizeof(buf) - 1, "playback_%u", chn+1);

	port = jack_port_register (client, buf,
		JACK_DEFAULT_AUDIO_TYPE,
		port_flags, 0);

	if (!port)
	{
	    printf( "jacknet_client: cannot register port for %s", buf);
	    break;
	}

	playback_ports = jack_slist_append (playback_ports, port);
    }
}


/*
 * The Sync callback... sync state is set elsewhere...
 * we will see if this is working correctly.
 * i dont really believe in it yet.
 */

int sync_cb( jack_transport_state_t state, jack_position_t *pos, void *arg ) {
    return sync_state;
}


/**
 * The process callback for this JACK application.
 * It is called by JACK at the appropriate times.
 */
int
process (jack_nframes_t nframes, void *arg)
{
    static int framecnt = 0;
    //static int old_count = 0;


    jack_position_t jack_pos;

    int rx_bufsize =  sizeof(jack_default_audio_sample_t) * capture_channels * nframes + sizeof(jacknet_packet_header);
    int tx_bufsize =  sizeof(jack_default_audio_sample_t) * playback_channels * nframes + sizeof(jacknet_packet_header);

    jack_default_audio_sample_t *buf;
    jack_port_t *port;
    JSList *node;
    channel_t chn;
    int size;

//    jack_default_audio_sample_t *out1 = 
//	(jack_default_audio_sample_t *) jack_port_get_buffer (output_port1, nframes);
//    jack_default_audio_sample_t *out2 = 
//	(jack_default_audio_sample_t *) jack_port_get_buffer (output_port2, nframes);
//    unsigned int *inputbuffer;
//    jack_default_audio_sample_t *in1 = 
//	(jack_default_audio_sample_t *) jack_port_get_buffer (input_port1, nframes);
//    jack_default_audio_sample_t *in2 = 
//	(jack_default_audio_sample_t *) jack_port_get_buffer (input_port2, nframes);
//    unsigned int *sendbuffer;


    unsigned int *packet_buf, *packet_bufX;

    // Allocate a buffer where both In and Out Buffer will fit
    packet_buf = alloca( (rx_bufsize>tx_bufsize) ? rx_bufsize : tx_bufsize );

    framecnt++;

    jacknet_packet_header *pkthdr = (jacknet_packet_header *)packet_buf;

    packet_bufX = packet_buf + sizeof(jacknet_packet_header) / sizeof(jack_default_audio_sample_t);

    //////////////// receive ////////////////
    ////////////////         ////////////////

    size = recv( outsockfd, packet_buf, rx_bufsize, MSG_DONTWAIT );
    
    // evaluate rcvd data.
    
    if( size == rx_bufsize ) {
	// packet has expected size
	chn = 0;
	node = capture_ports;
	while ( node != NULL)
	{
		int i;
		unsigned int val;
		port = (jack_port_t *) node->data;
		buf = jack_port_get_buffer (port, nframes);
	
		for (i = 0; i < nframes; i++)
		{
			unsigned int *ptr = (unsigned int *)&packet_bufX[i];
			val = ntohl( *ptr);
			buf[i] = *(float *)&val;
		}
		packet_bufX = (packet_bufX + nframes);
		node = jack_slist_next (node);
		chn++;
	}

	// Now evaluate packet header;
	if( sync_state != pkthdr->sync_state ) printf( "sync = %d\n", pkthdr->sync_state );
	sync_state = pkthdr->sync_state;
    }
    else
    {
	printf( "no: bufsize=%d, size=%d framecnt=%d\n", rx_bufsize, size, framecnt );
	chn = 0;
	node = capture_ports;
	while ( node != NULL)
	{
		int i;
		port = (jack_port_t *) node->data;
		buf = jack_port_get_buffer (port, nframes);
	
		for (i = 0; i < nframes; i++)
		{
			buf[i] = 0.0;
		}
		node = jack_slist_next (node);
		chn++;
	}
    }
    ////////////////
    ////////////////

    // reset packet_bufX... and then send...

    packet_bufX = packet_buf + sizeof(jacknet_packet_header) / sizeof(jack_default_audio_sample_t);

    //////////////// send ////////////////
    ////////////////      ////////////////
    chn = 0;
    node = playback_ports;
    while ( node != NULL)
    {
	int i;
	unsigned int val;
	port = (jack_port_t *) node->data;
	buf = jack_port_get_buffer (port, nframes);

	for (i=0; i < nframes; i++) {
	    unsigned int *ptr = (unsigned int *)&buf[i];
	    val = htonl( *ptr);
	    packet_bufX[i]=val;
	}

	packet_bufX = (packet_bufX + nframes);
	node = jack_slist_next (node);
	chn++;
    }

    // fill in packet hdr
    
    pkthdr->transport_state = jack_transport_query( client, &pkthdr->jack_pos );
    
    pkthdr->sample_rate = jack_get_sample_rate( client );
    
    int err = sendto( outsockfd, packet_buf, tx_bufsize, 0, &destaddr, sizeof( destaddr ) );
    return 0;      
}

/**
 * This is the shutdown callback for this JACK application.
 * It is called by JACK if the server ever shuts down or
 * decides to disconnect the client.
 */
void
jack_shutdown (void *arg)
{

	exit (1);
}

void
init_sockaddr_in ( struct sockaddr_in *name , const char *hostname , uint16_t port )
{
  name->sin_family = AF_INET ;
  name->sin_port = htons ( port ) ;
  if ( hostname ) {
    struct hostent *hostinfo = gethostbyname ( hostname ) ;
    if (hostinfo == NULL) {
      fprintf ( stderr , "init_sockaddr_in: unknown host: %s.\n" , hostname ) ;
    }
    name->sin_addr = *(struct in_addr *) hostinfo->h_addr ;
  } else {
    name->sin_addr.s_addr = htonl ( INADDR_ANY ) ;
  }
}

void 
printUsage()
{
fprintf(stderr, "usage: net_source [-n <jack name>] [-s <socket>] [-C <num channels>] [-P <num channels>] -p <host peer>\n"
		"\n"
		"  -n <jack name> - reports a different name to jack\n"
		"  -s <socket> select another socket than the default (3000).\n"
		"  -p <host peer> the hostname of the \"other\" machine running the jack-slave.\n"
		"  -P <num channels> number of playback channels.\n"
		"  -C <num channels> number of capture channels.\n"
		"\n");
}

int
main (int argc, char *argv[])
{
    const char **ports;
    char jack_name[30] = "net_source";
    char *peer_ip;
    int peer_socket = 3000;

    if (argc < 3) {
	printUsage();
	return 1;
    }
    extern char *optarg;
    extern int optind, optopt;
    int errflg=0;
    int c;

    while ((c = getopt(argc, argv, ":n:p:s:C:P:")) != -1) {
	switch(c) {
	    case 'n':
		strcpy(jack_name,optarg);
		break;
	    case 'p':
		peer_ip = optarg;
		break;
	    case 's':
		peer_socket = atoi(optarg);
		break;
	    case 'P':
		playback_channels = atoi(optarg);
		break;
	    case 'C':
		capture_channels = atoi(optarg);
		break;
	    case ':':       /* -n or -p without operand */
		fprintf(stderr,
			"Option -%c requires an operand\n", optopt);
		errflg++;
		break;
	    case '?':
		fprintf(stderr,
			"Unrecognized option: -%c\n", optopt);
		errflg++;
	}
    }
    if (errflg) {
	printUsage();
	exit(2);
    }

    outsockfd = socket( PF_INET, SOCK_DGRAM, 0 );
    insockfd = socket( PF_INET, SOCK_DGRAM, 0 );
    init_sockaddr_in( (struct sockaddr_in *)&destaddr, peer_ip, peer_socket );
    init_sockaddr_in( (struct sockaddr_in *)&bindaddr, NULL, peer_socket );
    bind( insockfd, &bindaddr, sizeof(bindaddr) );
    /* try to become a client of the JACK server */

    if ((client = jack_client_new (jack_name)) == 0) {
	fprintf (stderr, "jack server not running?\n");
	return 1;
    }

    /* 
       send a ping to the peer 
       -- this needs to be made more robust --
       */

    sendto( outsockfd, "x", 1, 0, &destaddr, sizeof( destaddr ) );

    /* tell the JACK server to call `process()' whenever
       there is work to be done.
       */

    jack_set_process_callback (client, process, 0);
    jack_set_sync_callback (client, sync_cb, 0);


    /* tell the JACK server to call `jack_shutdown()' if
       it ever shuts down, either entirely, or if it
       just decides to stop calling us.
       */

    jack_on_shutdown (client, jack_shutdown, 0);


    /* display the current sample rate. 
    */

    printf ("engine sample rate: %" PRIu32 "\n",
	    jack_get_sample_rate (client));

    /* create two (later n) ports */

    /* TODO: evaluate channel numbers and initialize in for(loop)
     *
     * needs: list< portname > :: iterator  portname_iter
     * 
     * input_port = jack_port_register (client, *( port_name_iter++ ), JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
     */

    alloc_ports( capture_channels, playback_channels );

    /* tell the JACK server that we are ready to roll */

    if (jack_activate (client)) {
	fprintf (stderr, "cannot activate client");
	return 1;
    }

    // Now sleep forever.......

    while(1)
	sleep(100);

    // Never reached. Well we will be a GtkApp someday....

    jack_client_close (client);
    exit (0);
}

